#include <Arduino.h>
#include "BTN.h"
#include "LDR.h"
#include "Dimmer.h"

Dimmer dimmer;

void Dimmer::Init(int relay, int led)
{
  m_relay = relay;
  m_led = led;
  pinMode(m_relay, OUTPUT);
  pinMode(m_led, OUTPUT);
  digitalWrite(m_led, LOW);
  m_bClientOff = m_bManualMode = m_bManualOff = false;
  // LOW is the default pin state
  TurnClientOff(kRelayClientOff == LOW);
}

void Dimmer::Configure()
{
  // Invoked if btn held when device turned on
  // loops until another btn press
  // applying the simple on/off logic to the LED so user can adjust threshhold by turning pot
  // blink the LED to say we're ready
  for (int i = 0; i < 20; i++)
  {
    TurnLedOff(i % 2);
    delay(50);
  }
  TurnLedOff(true);
  delay(100);
  while (!btn.CheckButtonPress())
    TurnLedOff(ShouldTurnOff(ldr.ReadAverage(), true));
  TurnLedOff(true);
}

void Dimmer::Start()
{
  m_LastCheckMS = millis();
  ResetRange();
}

void Dimmer::Loop()
{
  // the main loop
  // monitors LDR and controls relay
  if (btn.CheckButtonPress())
  {
    // button-press means manual over-ride to the other state
    Blink();
    m_bManualMode = true;
    m_bManualOff = !m_bClientOff;
    TurnClientOff(m_bManualOff);
    ResetRange();
#ifndef ARDUINO_AVR_TRINKET3
    Serial.print("Manual client off:"); Serial.println(m_bManualOff);
#endif    
  }
  else
  {
    unsigned long NowMS = millis();
    if ((NowMS - m_LastCheckMS) > kCheckAmbientIntervalMS)
    {
      m_LastCheckMS = NowMS;
      int reading = ldr.ReadAverage();
      UpdateRange(reading);
      // Determine what ambient light logic says the state should be:
      bool off = ShouldTurnOff(reading, false);
      if (m_bManualMode)
      {
        // if we would now match manual mode, turn manual over-ride off
        if (off == m_bManualOff)
        {
          m_bManualMode = false;
#ifndef ARDUINO_AVR_TRINKET3
          Serial.print("Manual disabled, client off:"); Serial.println(off);
#endif          
          if (TurnClientOff(off))
            ResetRange();
        }
      }
      else
      {
        if (TurnClientOff(off))
          ResetRange();
      }
    }
  } 
}

void Dimmer::Blink()
{
  TurnLedOff(false);
  delay(50);
  TurnLedOff(true);
}

bool Dimmer::ShouldTurnOff(int ldr, bool immediate)
{
  if (immediate)  // no special logic
    return ldr < kAmbientThreshold;
  
  // more detailed logic, avoid rapid fluctuations is state
  // The client is turned off or on if the ldr reading is below or above the threshold respectively
  // Howver, there is a "grey area" kAmbientThreshold +/- kAmbientGreyArea
  // Transitions within the grey area only happen if enough time (kGreyAreaTimoutMS) has passed since the last one
  unsigned long NowMS = millis();
  if (m_bClientOff)
  {
    if (m_OffMin <= (kAmbientThreshold - kAmbientGreyArea) || m_OffMin == 1024)
    {
      // ambient dipped below grey area since transition, can switch state now
      return ldr < kAmbientThreshold;
    }
    else if ((NowMS - m_LastClientStateChangeMS) > kGreyAreaTimoutMS)
    {
      // ambient did not go below grey area since last transition, can switch state if it's been long enough
      return ldr < kAmbientThreshold;
    }
  }
  else
  {
    if (m_OnMax >= (kAmbientThreshold + kAmbientGreyArea) || m_OnMax == 0)
    {
      // ambient went above grey area since transition, can switch state now
      return ldr < kAmbientThreshold;
    }
    else if ((NowMS - m_LastClientStateChangeMS) > kGreyAreaTimoutMS)
    {
      // ambient did not go above grey area since last transition, can switch state if it's been long enough
      return ldr < kAmbientThreshold;
    }
  }
  return m_bClientOff;
}

bool Dimmer::TurnClientOff(bool off)
{
  // turns client on/off via relay, returns true if client state changed
  if (m_bClientOff != off)
  {
    m_bClientOff = off;
#ifndef ARDUINO_AVR_TRINKET3 
    Serial.print("TurnClientOff:"); Serial.println(m_bClientOff);
#endif    
    digitalWrite(m_relay, m_bClientOff?kRelayClientOff:!kRelayClientOff);
    m_LastClientStateChangeMS = millis();
    return true;
  }
  return false;
}

void Dimmer::TurnLedOff(bool off)
{
  digitalWrite(m_led, off?LOW:HIGH);
}

void Dimmer::ResetRange()
{
  m_OnMax = 0;
  m_OffMin = 1024;
//Serial.println("Min/Max reset");
}

void Dimmer::UpdateRange(int ldr)
{
  if (m_bClientOff)
    m_OffMin = min(m_OffMin, ldr);
  else
    m_OnMax = max(m_OnMax, ldr);
//Serial.print("Min:");Serial.print(m_OffMin);Serial.print("  Max:");Serial.println(m_OnMax);
}
